from datetime import datetime, time

from flask import request, jsonify, abort, current_app
from permissive_dict import PermissiveDict


class FlaskSerializeMixin:
    """
    Base mix in class to implement serialization and update methods for use
    with Flask and SQL Alchemy models
    """
    # fields that should not be exposed by serialization
    exclude_serialize_fields = []
    exclude_json_serialize_fields = []
    # list of fields to be used when updating
    update_fields = []
    # list of fields used when creating
    create_fields = []
    # order by ascending field
    order_by_field = None
    order_by_field_desc = None
    # default error code
    __http_error_code = 400
    # list of fields used to set the update date/time
    timestamp_fields = ['updated', 'timestamp']
    # method to be used to timestamp with update_timestamp method
    timestamp_stamper = datetime.utcnow
    # list of property names that are relationships to be included in serialization
    relationship_fields = []
    # add your own converters here
    column_type_converters = {}
    # add or replace conversion types
    convert_types = [{'type': bool, 'method': lambda v: 'y' if v else 'n'},
                     {'type': bytes, 'method': lambda v: v.encode()}]
    __json_types = [str, dict, list, int, float, bool]
    # default field name when restricting to a particular user
    fs_user_field = 'user'
    # properties or fields to return when updating using get or post
    update_properties = []
    # db is required to be set for updating/deletion functions
    db = None
    # cache model properties
    __model_props = {}
    # previous values of an instance before update attempted
    previous_field_value = {}
    # current version
    __version__ = '1.5.0'

    def before_update(self, data_dict):
        """
        hook to call before any update_from_dict so that you may alter update values
        before the item is written in preparation for update to db

        :param data_dict: the new data to apply to the item
        :return: the new data_dict to use for updating
        """
        return data_dict

    def to_date_short(self, d):
        """
        convert the given date field to a short date / time without fractional seconds

        :param d: {datetime.datetime} the value to convert
        :return: {String} the short date
        """
        return str(d).split('.')[0]

    @classmethod
    def query_by_access(cls, user=None, **kwargs):
        """
        filter a query result to that optionally owned by the given user and
        can_access()

        :param query: SQLAlchemy query
        :param user: the user to use as a filter, default relationship name is user (fs_user_field)
        :return: a list of query results
        """

        if user:
            kwargs[cls.fs_user_field] = user

        query = cls.query.filter_by(**kwargs)
        items = [item for item in query if item.can_access()]

        return items

    @classmethod
    def get_by_user_or_404(cls, item_id, user=None):
        """
        return the object with the given primary key id that is optionally owned by the given user and
        can_access()

        :param item_id: object primary key, id
        :param user: the user to use as a filter, default relationship name is user (fs_user_field)
        :return: the object
        :throws: 404 exception if not found
        """

        kwargs = {'id': item_id}
        if user:
            kwargs[cls.fs_user_field] = user
        item = cls.query.filter_by(**kwargs).first_or_404()
        if not item.can_access():
            abort(404)
        return item

    @classmethod
    def json_filter_by(cls, **kwargs):
        """
        return a list in json format using the filter_by arguments

        :param kwargs: SQLAlchemy query.filter_by arguments
        :return: flask response with json list of results
        """
        results = cls.query.filter_by(**kwargs)
        return cls.json_list(results)

    @classmethod
    def json_get(cls, item_id):
        """
        return an item in json format using the item_id as a primary key

        :param item_id: {primary key} the primary key of the item to get
        :return: flask response with json item, or {} if not found or no access
        """
        item = cls.query.get(item_id)
        if not item or not item.can_access():
            return jsonify({})
        return item.as_json

    @classmethod
    def json_list(cls, query_result, prop_filters=None):
        """
        Return a list in json format from the query_result.
        When order_by_field is defined sort by that field in ascending order.
        Only returns those that can_access()

        :param query_result: sql alchemy query result
        :param prop_filters: dictionary of filter elements to restrict results
        :return: flask response with json list of results
        """
        # ascending
        items = [item.__as_exclude_json_dict() for item in query_result if item.can_access()]

        if len(items) > 0 and cls.order_by_field:
            items = sorted(items, key=lambda i: i[cls.order_by_field])

        # descending
        if len(items) > 0 and cls.order_by_field_desc:
            items = sorted(items, key=lambda i: i[cls.order_by_field_desc], reverse=True)

        if prop_filters:
            filtered_result = []
            for item in items:
                for k, v in prop_filters.items():
                    if item.get(k) == v:
                        filtered_result.append(item)
                        break
            items = filtered_result

        return jsonify(items)

    @classmethod
    def dict_list(cls, query_result):
        """
        return a list of dictionary objects from the sql query result
        without exclude_serialize_fields fields
        for only those than can_access()

        :param query_result: sql alchemy query result
        :return: list of dict objects
        """
        return [item.__as_exclude_json_dict() for item in query_result if item.can_access()]

    @property
    def as_json(self):
        """
        the sql object as a json response without the excluded fields

        :return: flask response json object
        """
        return jsonify(self.__as_exclude_json_dict())

    def fs_private_field(self, field_name):
        return False

    def __as_exclude_json_dict(self):
        """
        private: get a dict that is used to serialize to web clients
        without fields in exclude_json_serialize_fields and exclude_serialize_fields
        excludes any private_field

        :return: dictionary
        """
        return {k: v for k, v in self.as_dict.items() if
                k not in self.exclude_json_serialize_fields}

    def property_converter(self, value):
        """
        convert to a json compatible format.

        * complex - just uses str conversion
        * datetime - short format as per to_date_short
        * set - becomes a list
        * anything else not supported by json becomes a str

        override or extend this method to alter defaults

        :param value: value to convert
        :return: the new value
        """
        if value is None:
            return None
        if isinstance(value, datetime):
            return self.to_date_short(value)
        if isinstance(value, set):
            return list(value)
        if isinstance(value, FlaskSerializeMixin):
            return value.as_dict
        if type(value) not in self.__json_types:
            return str(value)
        return value

    @staticmethod
    def __relationship_converter(relationships):
        """
        convert a child SQLAlchemy result set into a python
        dictionary list.

        :param relationships: SQLAlchemy result set
        :return: list of dict objects
        """
        if isinstance(relationships, FlaskSerializeMixin):
            return relationships.as_dict
        return [item.as_dict for item in relationships]

    @staticmethod
    def __lob_converter(value):
        """
        convert a LOB into a decoded string

        :param value: lob to convert
        :return: decoded string
        """
        if not value:
            return ''
        value = value.decode()
        return value

    @staticmethod
    def __sqlite_date_converter(value):
        """
        convert an ISO-9 date or datetime from a form etc into a datetime as sqlite does
        not handle date conversions nicely

        :param value: form / json data to convert
        :return: corrected datetime or time
        """
        if isinstance(value, datetime) or isinstance(value, time):
            return value
        # assumes ISO 8601 as per javascript standard for dates
        try:
            return datetime.strptime(value, '%Y-%m-%dT%H:%M')
        except:
            return datetime.strptime(value, '%Y-%m-%d')

    def __get_props(self):
        """
        get the properties for this table to be used for introspection

        :return: properties EasyDict
        """
        props = self.__model_props.get(self.__table__)
        if not props:
            props = PermissiveDict(name=self.__table__.name, id=0, primary_key_field='id')
            props.converters = {'DATETIME': self.to_date_short,
                                'PROPERTY': self.property_converter,
                                'RELATIONSHIP': self.__relationship_converter,
                                'NUMERIC': float,
                                'DECIMAL': float,
                                'LOB': self.__lob_converter,
                                'BLOB': self.__lob_converter,
                                'CLOB': self.__lob_converter,
                                }

            # SQL columns
            props.__exclude_fields = ['as_dict', 'as_json'] + self.exclude_serialize_fields
            field_list = self.__table__.columns
            if 'sqlite' in self.__table__.dialect_options:
                props.DIALECT = 'sqlite'
                self.convert_types.append(
                    {'type': datetime, 'method': self.__sqlite_date_converter}),
            # detect primary field
            for f in field_list:
                if f.primary_key:
                    props.id = str(getattr(self, f.name, ''))
                    props.primary_key_field = f.name
                    break

            # add class properties
            field_list += [PermissiveDict(name=p, type='PROPERTY') for p in dir(self.__class__) if
                           isinstance(getattr(self.__class__, p), property)]
            # add relationships
            field_list += [PermissiveDict(name=p, type='RELATIONSHIP') for p in self.relationship_fields]
            # add custom converters
            for converter, method in self.column_type_converters.items():
                props.converters[converter] = method
            # exclude fields / props
            props.field_list = []
            for f in field_list:
                if f.name not in props.__exclude_fields:
                    f.c_type = str(f.type).split('(')[0]
                    f.converter = props.converters.get(f.c_type)
                    if not f.converter:
                        # any non json supported types gets a str
                        if getattr(f.type, 'python_type', None) not in self.__json_types:
                            f.converter = str
                    props.field_list.append(f)

            self.__model_props[self.__table__] = props
        return props

    def __get_fields(self):
        """
        return a list of field objects that are valid
        using fs_private_field
        [{name,c__type,converter},...]

        :return:
        """
        fields = []
        for c in self.__get_props().field_list:
            if not self.fs_private_field(c.name):
                fields.append(c)
        return fields

    @property
    def as_dict(self):
        """
        convert a sql alchemy query result item to dict
        override these properties to control the result:

        * relationship_fields - follow relationships listed
        * exclude_serialize_fields - exclude listed fields from serialization
        * column_type_converters - add additional sql column type converters to DATETIME, PROPERTY and RELATIONSHIP
        :return {dictionary} the item as a dict
        """

        # built in converters
        # can be replaced using column_type_converters
        d = {}

        for c in self.__get_fields():
            try:
                d[c.name] = v = getattr(self, c.name, '')
            except Exception as e:
                v = str(e)

            if v is None:
                d[c.name] = ''
            elif c.converter:
                try:
                    d[c.name] = c.converter(v)
                except Exception as e:
                    d[c.name] = 'Error:"{}". Failed to convert [{}] type:{}'.format(e, c.name, c.c_type)
        return d

    def __json_api_dict(self):
        """
        start of returning as JSON_API.  Unused

        :return: {id, name, attributes(dict)}
        """
        props = self.__get_props()
        d = dict(id=props.id, type=props.name, attributes=self.as_dict)
        return jsonify(d)

    def __json_api_list(self, query_result):
        """
        start of returning as JSON_API.  Unused
        returns a list of items where can_access() from query_result

        :param query_result:
        :return: list of {id, name, attributes(dict)}
        """
        dict_list = self.dict_list(query_result=query_result)
        d = dict(data=[item.__json_api_dict() for item in dict_list])
        return jsonify(d)

    def __get_update_field_type(self, field, value):
        """
        get the type of the update to db field from cached table properties

        :param field:
        :return: class of the type
        """
        props = self.__get_props()
        if props:
            for f in props.field_list:
                if f.name == field:
                    if f.c_type.startswith("VARCHAR") or f.c_type.startswith("CHAR") or f.c_type.startswith("TEXT"):
                        return str
                    if f.c_type.startswith("INTEGER"):
                        return int
                    if f.c_type.startswith("FLOAT") or f.c_type.startswith("REAL") or f.c_type.startswith("NUMERIC"):
                        return float
                    if f.c_type.startswith("DATE") or f.c_type.startswith("TIME"):
                        return datetime
                    if f.c_type.startswith("BOOLEAN"):
                        return bool
                    if 'LOB' in f.c_type:
                        return bytes

        return None

    def __convert_value(self, name, value):
        """
        convert the value based upon type to a representation suitable for saving to the db
        override built in conversions by setting the value of convert_types.
        First uses bare value to determine type and then uses db derived values
        ie:
        convert_types = [{'type':bool, 'method': lambda x: not x}]

        :param name: name of the field to update
        :param value: value to update with
        :return: the converted value
        """
        for t in self.convert_types:
            if isinstance(value, t['type']):
                value = t['method'](value)
                return value

        instance_type = self.__get_update_field_type(name, value)
        if instance_type:
            for t in self.convert_types:
                if instance_type == t['type']:
                    value = t['method'](value)
                    return value
        return value

    @classmethod
    def request_create_form(cls, **kwargs):
        """
        create a new item from a form in the current request object
        throws error if something wrong. Use **kwargs to set the object properties of the newly created item.

        :return: the new created item
        """
        new_item = cls(**kwargs)

        create_fields = list(new_item.create_fields)
        if len(create_fields or '') == 0:
            create_fields = [c.name for c in new_item.__get_fields() if
                             isinstance(c, cls.db.Column) and c.name != new_item.__get_props().primary_key_field]

        try:
            json_data = request.get_json(force=True)
            if len(json_data) > 0:
                for field in create_fields:
                    if cls.db and isinstance(field, cls.db.Column):
                        field = field.name
                    if field in json_data:
                        value = json_data.get(field)
                        setattr(new_item, field, new_item.__convert_value(field, value))
        except:
            for field in create_fields:
                if cls.db and isinstance(field, cls.db.Column):
                    field = field.name
                if field in request.form:
                    value = request.form.get(field)
                    setattr(new_item, field, new_item.__convert_value(field, value))

        new_item.verify(create=True)
        new_item.update_timestamp()
        cls.db.session.add(new_item)
        cls.db.session.commit()
        return new_item

    def request_update_form(self):
        """
        update/create the item using form data from the request object
        only present fields are updated
        throws error if validation or can_update() fails

        :return: True when complete
        """
        if request.content_type == 'application/json' or request.method == 'PUT':
            return self.request_update_json()
        else:
            self.update_from_dict(request.form)
        if not self.can_update():
            return False
        self.verify()
        self.update_timestamp()
        if not self.db:
            raise Exception('FlaskSerializeMixin property "db" is not set')
        self.db.session.add(self)
        self.db.session.commit()
        return True

    def request_update_json(self):
        """
        Update an item from request json data or PUT params, probably from a PUT or PATCH.
        Throws exception if not valid or can_update() fails

        :return: True if item updated
        """

        if not self.can_update():
            return False
        try:
            json_data = request.get_json(force=True)
        except Exception as e:
            json_data = request.values
            if len(json_data) == 0:
                current_app.logger.exception(e)

        self.update_from_dict(json_data)

        self.verify()
        self.update_timestamp()
        if not self.db:
            raise Exception('FlaskSerializeMixin property "db" is not set')
        self.db.session.add(self)
        self.db.session.commit()
        return True

    def update_timestamp(self):
        """
        update any timestamp fields using the Class timestamp method if those fields exist

        """
        for field in self.timestamp_fields:
            if hasattr(self, field):
                setattr(self, field, self.timestamp_stamper())

    def update_from_dict(self, data_dict):
        """
        uses a dict to update fields of the model instance.  sets previous values to
        self.previous_values[field_name] before the update

        :param data_dict: the data to update
        :return:
        """
        data_dict = self.before_update(data_dict)
        update_fields = list(self.update_fields)
        if len(self.update_fields or '') == 0:
            update_fields = [c.name for c in self.__get_fields() if
                             isinstance(c, self.db.Column) and c.name != self.__get_props().primary_key_field]
        for field in update_fields:
            if self.db and isinstance(field, self.db.Column):
                field = field.name
            self.previous_field_value[field] = getattr(self, field)
            if field in data_dict:
                setattr(self, field, self.__convert_value(field, data_dict[field]))

    def can_access(self):
        """
        return True if allowed to access, return false if not

        :return: True/False
        """
        return True

    def can_update(self):
        """
        return True if allowed to update, raise an error if not allowed

        :return: True if can update
        """
        return self.can_access()

    def can_delete(self):
        """
        raise an exception if deletion is not allowed or True
        if deletion is allowed.  Default is can_update() returns false
        aborts with 403

        :return: True if can delete
        """
        if not self.can_update():
            abort(403)
        return True

    def verify(self, create=False):
        """
        raise exception if item is not valid for put/patch/post

        :param: create - True if verification is for a new item
        """
        pass

    def __return_properties(self):
        """
        when returning success codes from a put/post update return a dict
        composed of the property values from the update_properties list.
        ie:
        return jsonify({'message': 'Updated', 'properties': item.__return_properties()})
        this can be used to communicate from the model on the server to the JavaScript code
        interesting things from updates, by default returns all accessible fields / properties.

        :return: dictionary of properties
        """
        return {prop: self.property_converter(getattr(self, prop)) for prop in self.update_properties}

    @classmethod
    def get_delete_put_post(cls, item_id=None, user=None, prop_filters=None):
        """
        get, delete, post, put with JSON/FORM a single model item

        * any access uses can_access() to check for accessibility
        * any update uses can_update() to check for update permission

        :param item_id: the primary key of the item - if none and method is 'GET' returns all items
        :param user: user to use as query filter.
        :param prop_filters: dictionary of key:value pairs to limit results to.
        :return: json object: {message}, or the item.  throws error when problem
        """
        item = None
        if user is not None and item_id is not None:
            item = cls.get_by_user_or_404(item_id, user=user)
        elif item_id is not None:
            item = cls.query.get_or_404(item_id)
            if not item.can_access():
                abort(404)
        elif request.method == 'GET':
            # no item id get a list of items
            if user:
                kwargs = {cls.fs_user_field: user}
                result = cls.query.filter_by(**kwargs)
            else:
                result = cls.query.all()
            return cls.json_list(result, prop_filters=prop_filters)

        try:
            if not item:
                if request.method == 'POST':
                    return cls.request_create_form().as_json
                abort(405)

            # get a single item
            if request.method == 'GET':
                return item.as_json

            elif request.method == 'POST' or request.method == 'PUT':
                # update single item
                item.request_update_form()
                return jsonify(
                    dict(message='Updated', item=item.__as_exclude_json_dict(), properties=item.__return_properties()))

            elif request.method == 'DELETE':
                # delete a single item
                item.can_delete()
                cls.db.session.delete(item)
                cls.db.session.commit()
                return jsonify(dict(item=item.as_dict, message='Deleted'))

        except Exception as e:
            return str(e), cls.__http_error_code

    @classmethod
    def json_first(cls, **kwargs):
        """
        return the first result in json response format using the filter_by arguments, or {} if no result
        or not can_access()

        :param kwargs: SQLAlchemy query.filter_by arguments
        :return: flask response json item or {} if no result
        """
        item = cls.query.filter_by(**kwargs).first()
        if not item or not item.can_access():
            return jsonify({})

        return item.as_json


def FlaskSerialize(db=None):
    """
    return the FlaskSerializeMixin mixin class, optionally initialize the db values

    :param db: (optional) SQLAlchemy db instance
    :return: FlaskSerializeMixin mixin
    """
    FlaskSerializeMixin.db = db
    return FlaskSerializeMixin



from flask import flash, redirect, url_for, request, render_template


class FormPageMixin:
    # form_page properties
    form_page_form = None  # name of the form to use
    form_page_route_update = form_page_route_create = ''  # method names for successful redirect
    form_page_template = ''  # template for editing
    # all these called with .format(cls) or .format(item) as appropriate
    form_page_update_format = 'Updated {}'
    form_page_create_format = 'Created {}'
    form_page_new_title_format = 'New {}'
    form_page_update_title_format = 'Edit {}'

    @classmethod
    def form_page(cls, item_id=None):
        """
        Do all the work for creating and editing items using a template and a wtf form.
        Prerequisites.

        cls.form = WTFFormClass
        cls.form_route_create = Name of the method to redirect after create, uses: url_for(cls.form_route_create, item_id=id)
        cls.form_route_update = Name of the method to redirect after updating, uses: url_for(cls.form_route_update, item_id=id)
        cls.form_template = 'Location of the template file to allow edit/add'

        WTFFormClass - needs to have a hidden id with the name 'id'

        :param item_id: Item_id if editing, otherwise None
        :return: templates and redirects as required
        """
        form = cls.form_page_form()
        if item_id:
            item = cls.query.get_or_404(item_id)
            title = cls.form_page_update_title_format.format(item)
        else:
            title = cls.form_page_new_title_format.format(cls)
            item = dict(id=None)

        if form.validate_on_submit():
            try:
                if item_id:
                    item.request_update_form()
                    msg = cls.form_page_update_format.format(item)
                    if msg:
                        flash(msg)
                    return redirect(url_for(cls.form_page_route_update, item_id=item_id))
                else:
                    new_item = cls.request_create_form()
                    msg = cls.form_page_create_format.format(new_item)
                    if msg:
                        flash(msg)
                    return redirect(url_for(cls.form_page_route_create, item_id=new_item.id))
            except Exception as e:
                flash(str(e), category='danger')
        elif request.method == 'GET':
            if not item_id:
                # new blank form
                item = cls()

            form = cls.form_page_form(obj=item)

        if form.errors:
            flash(''.join([f'{form[f].label.text}: {"".join(e)} ' for f, e in form.errors.items()]), category='danger')

        return render_template(cls.form_page_template, title=title, item_id=item_id, item=item, form=form)



import os
import random
import string
import time
from datetime import datetime, timedelta

from flask import Flask, redirect, url_for, Response, request, render_template, flash
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, ValidationError, validators, DecimalField, SubmitField

from flask_serialize.flask_serialize import FlaskSerialize
from flask_serialize.form_page import FormPageMixin

app = Flask("test_app")
app.testing = True
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
app.config['SQLALCHEMY_DATABASE_URI'] = os.environ.get("SQLALCHEMY_DATABASE_URI", "sqlite:///:memory:")
db = SQLAlchemy(app)
migrate = Migrate(app, db)
fs_mixin = FlaskSerialize(db)


# =========================
# TINY TEST FLASK APP
# =========================


def random_string(length=20):
    """
    return a <length> long character random string of ascii_letters
    :param length: {int} number of characters to return
    :return:
    """
    return ''.join(random.sample(string.ascii_letters, length))


class EditForm(FlaskForm):
    setting_type = StringField('setting_type', [validators.DataRequired()])
    key = StringField('key')
    value = StringField('value')
    number = IntegerField('number')
    deci = DecimalField('deci')
    submit = SubmitField()


@app.route('/')
@app.route('/<item_id>')
def page_index(item_id=None):
    settings = Setting.query.all()
    return render_template("index.html", title='Flask Serialize Tester', settings=settings, key=random_string(),
                           setting_type=random_string(), value=random_string())


# Get all items as a json list.
# post, put update
# and get single
@app.route('/setting_post', methods=['POST', 'PUT'])
@app.route('/setting_get_all', methods=['GET'])
@app.route('/setting_post/<int:item_id>', methods=['POST'])
@app.route('/setting_put/<int:item_id>', methods=['PUT'])
@app.route('/setting_get/<int:item_id>', methods=['GET'])
@app.route('/setting_user/<user>', methods=['GET'])
@app.route('/setting_id_user/<int:item_id>/<user>', methods=['GET'])
def route_setting_get_delete_put_post(item_id=None, user='Andrew'):
    key = request.args.get('key')
    if key and request.method == 'GET':
        return Setting.get_delete_put_post(item_id, prop_filters={"key": key})
    return Setting.get_delete_put_post(item_id, user)


@app.route('/sub_setting_delete/<int:item_id>', methods=['DELETE'])
@app.route('/sub_setting_put/<int:item_id>', methods=['PUT', 'POST'])
@app.route('/sub_setting_get/<int:item_id>', methods=['GET'])
def route_sub_setting_get_delete_put_post(item_id=None, user='fake'):
    return SubSetting.get_delete_put_post(item_id, user)


@app.route('/simple_add', methods=['POST'])
@app.route('/simple_edit/<int:item_id>', methods=['PUT', 'POST'])
def route_simple_get_delete_put_post(item_id=None, user=None):
    return SimpleModel.get_delete_put_post(item_id, user)


@app.route('/setting_get_json/<int:item_id>', methods=['GET'])
def route_setting_get_json(item_id):
    return Setting.json_get(item_id)


@app.route('/setting_json_first/<key>', methods=['GET'])
def route_setting_json_first(key):
    return Setting.json_first(key=key)


@app.route('/setting_get_key/<key>', methods=['GET'])
def route_setting_get_key(key):
    """
    get the first item that matches by a setting key

    :param key:
    :return:
    """
    return Setting.query.filter_by(key=key).first().as_json


@app.route('/setting_json_api/<int:id>', methods=['GET'])
def route_setting_get_json_api_key(id):
    """
    get the first item that matches by a setting key

    :param key:
    :return:
    """
    item = Setting.query.get(id)
    return item.json_api()


# Delete a single item.

@app.route('/setting_delete/<int:item_id>', methods=['DELETE'])
def route_setting_delete(item_id):
    return Setting.get_delete_put_post(item_id)


@app.route('/setting_update/<int:item_id>', methods=['PUT'])
def route_setting_update(item_id):
    """
    update from a json object

    :param item_id: item to update
    :return:
    """
    item = Setting.query.get_or_404(item_id)
    try:
        item.request_update_json()
    except Exception as e:
        print(e)
        return Response('Error updating item: ' + str(e), 500)
    return 'Updated'


@app.route('/sub_setting_add/<int:setting_id>', methods=['POST'])
def route_sub_setting_add(setting_id):
    """
    add as a child using kwargs on request create form

    :param setting_id:
    :return:
    """
    setting = Setting.query.get_or_404(setting_id)
    try:
        SubSetting.request_create_form(setting_id=setting.id)
    except Exception as e:
        return str(e), 500
    return redirect(url_for("route_setting_edit_add", item_id=setting_id))


@app.route('/setting_edit/<int:item_id>', methods=['POST', 'GET'])
@app.route('/setting_add', methods=['POST'])
def route_setting_edit_add(item_id=None):
    if item_id:
        item = Setting.query.get_or_404(item_id)
    else:
        item = {}
    form = EditForm(obj=item)

    if form.validate_on_submit():
        if item_id:
            try:
                item.request_update_form()
            except Exception as e:
                print(e)
                return Response('Error updating item: ' + str(e), 500)
            return redirect(url_for('route_setting_edit_add', item_id=item_id))
        else:
            try:
                Setting.request_create_form()
                return redirect(url_for('page_index'))
            except Exception as e:
                print(e)
                return Response('Error creating item: ' + str(e), 500)
    for err in form.errors:
        flash('**form error** {} {}'.format(str(err), form.errors[err]))

    return render_template("setting_edit.html", item=item, title='Edit Setting', form=form)


@app.route('/setting_form_edit/<int:item_id>', methods=['POST', 'GET'])
@app.route('/setting_form_add', methods=['POST'])
def route_setting_form(item_id=None):
    new_item = Setting.form_page(item_id)
    return new_item


@app.route('/datetest', methods=['POST'])
@app.route('/datetest/<int:item_id>', methods=['PUT'])
def route_datetest(item_id=None):
    return DateTest.get_delete_put_post(item_id)


# =========================
# MODELS
# =========================


class SubSetting(fs_mixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    created = db.Column(db.DateTime, default=datetime.utcnow)
    sub_updated = db.Column(db.DateTime, default=datetime.utcnow)

    setting_id = db.Column(db.Integer, db.ForeignKey('setting.id'))

    fake_user = db.Column(db.String(120), index=True, default='fake')
    flong = db.Column(db.String(120), index=True, default='flang')
    boolean = db.Column(db.Boolean, default=True)

    fs_user_field = 'fake_user'
    update_properties = create_fields = update_fields = ['flong', 'boolean']
    convert_types = [{'type': bool, 'method': lambda v: (type(v) == bool and v) or str(v).lower() == 'true'},
                     ]

    @staticmethod
    def one_day_ago():
        return datetime.utcnow() - timedelta(days=1)

    def to_date_short(self, date_value):
        """
        override DATETIME conversion behaviour to return unix time

        :param date_value:
        :return:
        """
        if not date_value:
            return 0

        return int(time.mktime(date_value.timetuple())) * 1000

    timestamp_fields = ['sub_updated']
    timestamp_stamper = one_day_ago


class Single(fs_mixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    wang = db.Column(db.String(120), default='wang')
    setting_id = db.Column(db.Integer, db.ForeignKey('setting.id'))


class DateTest(fs_mixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    a_date = db.Column(db.DateTime, default=datetime.utcnow)
    update_fields = create_fields = ['a_date']


class Setting(fs_mixin, FormPageMixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)

    setting_type = db.Column(db.String(120), index=True, default='misc')
    key = db.Column(db.String(120), index=True)
    value = db.Column(db.String(3000), default='')
    number = db.Column(db.Integer, default=0)
    active = db.Column(db.String(1), default='y')
    created = db.Column(db.DateTime, default=datetime.utcnow)
    updated = db.Column(db.DateTime, default=datetime.utcnow)
    scheduled = db.Column(db.DateTime, default=datetime.utcnow)
    user = db.Column(db.String(10), default='Andrew')
    floaty = db.Column(db.Float(), default=1.1)
    deci = db.Column(db.Numeric(2), default=0.0)
    # converter fields
    lob = db.Column(db.LargeBinary())
    # relationships
    sub_settings = db.relationship('SubSetting', backref='setting', cascade="all, delete-orphan")
    single = db.relationship('Single', backref='setting', uselist=False, cascade="all, delete-orphan")

    # serializer fields
    update_fields = [setting_type, 'value', 'key', 'active', 'number', 'floaty', 'scheduled', 'deci', 'lob']
    create_fields = [setting_type, 'value', 'key', 'active', 'user', 'floaty', 'number', 'deci', 'lob']
    exclude_serialize_fields = ['created']
    exclude_json_serialize_fields = ['updated']
    relationship_fields = ['sub_settings', 'single']
    update_properties = ['prop_test']
    order_by_field = 'value'
    # lob
    column_type_converters = {'LOB': lambda v: str(v)}
    # convert types
    scheduled_date_format = "%Y-%m-%d %H:%M:%S"
    convert_types = [
        {'type': bool, 'method': lambda v: 'y' if (type(v) == bool and v) or str(v).lower() == 'true' else 'n'},
        {'type': int, 'method': lambda n: int(n) * 2},
        {'type': datetime, 'method': lambda n: datetime.strptime(n, Setting.scheduled_date_format)},
        {'type': bytes, 'method': lambda v: v.encode()}
    ]
    # form_page
    form_page_form = EditForm
    form_page_route_update = 'route_setting_form'
    form_page_route_create = 'page_index'
    form_page_template = 'setting_edit.html'
    form_page_new_title_format = 'New Setting'

    @property
    def last_sub_setting(self):
        if len(self.sub_settings) > 0:
            return self.sub_settings[-1]
        else:
            return None

    def before_update(self, data_dict):
        d = dict(data_dict)
        d['active'] = d.get('active', 'n')
        return d

    # checks if Flask-Serialize can access
    def can_access(self):
        if self.value == '123456789':
            return False
        return True

    # checks if Flask-Serialize can access
    def can_update(self):
        if not self.can_access():
            raise Exception('Update not allowed.  Magic value!')
        return True

    # checks if Flask-Serialize can delete
    def can_delete(self):
        if self.value == '1234':
            raise Exception('Deletion not allowed.  Magic value!')

    # checks if Flask-Serialize can create/update
    def verify(self, create=False):
        if len(self.key or '') < 1:
            raise ValidationError('Missing key')

        if self.value == '666':
            raise ValidationError('Value is Devils Number')

        if len(self.setting_type or '') < 2:
            raise ValidationError('Insufficient setting type')

        if create:
            self.add_single('flang')

    def fs_private_field(self, field_name):
        if field_name.upper() == 'KEY' and self.key == 'private':
            return True
        return False

    @property
    def prop_test(self):
        return 'prop:' + self.value

    @property
    def prop_complex(self):
        return complex(self.number, self.number * 2)

    @property
    def prop_error(self):
        return 'prop:' + str(1 / 0)

    @property
    def prop_test_dict(self):
        return {'prop': self.value}

    @property
    def prop_datetime(self):
        return self.created

    @property
    def prop_set(self):
        return set([1, 2, 3, 'four'])

    def add_sub(self, flong):
        sub = SubSetting(setting=self, flong=flong)
        db.session.add(sub)
        db.session.commit()
        return sub

    def add_single(self, flong):
        sing = Single(setting=self, wang=flong)
        db.session.add(sing)
        return sing

    def __repr__(self):
        return 'Setting {}'.format(self.key)


class SimpleModel(fs_mixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    value = db.Column(db.String(30), default='')

    @property
    def prop(self):
        return 'prop:' + self.value

    def __repr__(self):
        return '<SimpleModel %r>' % (self.value)


if __name__ == '__main__':
    app.config['TESTING'] = True
    app.config['SECRET_KEY'] = 'Testing'
    app.config['WTF_CSRF_ENABLED'] = False
    db.create_all()
    app.run(debug=True, port=5055)



import json
import random
import string
import time
from datetime import datetime

import pytest

from test.test_flask_app import app, db, Setting, SubSetting, SimpleModel, DateTest


def random_string(length=20):
    """
    return a <length> long character random string of ascii_letters
    :param length: {int} number of characters to return
    :return:
    """
    return ''.join(random.sample(string.ascii_letters, length))


# =========================
# TESTS
# =========================

@pytest.fixture
def client():
    app.config['TESTING'] = True
    app.config['SECRET_KEY'] = 'Testing'
    app.config['WTF_CSRF_ENABLED'] = False
    client = app.test_client()

    with app.app_context():
        db.create_all()

    yield client

    db.drop_all()


def add_setting(client, key=random_string(), value='test-value', number=0):
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value, number=number))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    return item


def test_order_by_field(client):
    # add plenty
    count = 10
    for z in range(count):
        add_setting(client, key=str(z), value=str(z + 1))

    Setting.order_by_field = 'id'
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    json_settings = rv.json
    assert len(json_settings) == count
    sorted_list = sorted(json_settings, key=lambda i: i['id'])
    for z in range(count):
        assert json_settings[z]['id'] == sorted_list[z]['id']
    # ascending.
    Setting.order_by_field = 'value'
    rv = client.get('/setting_get_all')
    json_settings = rv.json
    sorted_list = sorted(json_settings, key=lambda i: i['value'])
    for z in range(count):
        assert json_settings[z]['value'] == sorted_list[z]['value']
    Setting.order_by_field = None
    Setting.order_by_field_desc = 'value'
    # descending
    rv = client.get('/setting_get_all')
    json_settings = rv.json
    sorted_list = sorted(json_settings, key=lambda i: i['value'], reverse=True)
    for z in range(count):
        assert json_settings[z]['value'] == sorted_list[z]['value']


def test_get_filter(client):
    key_1 = random_string()
    key_2 = random_string()
    # test add
    add_setting(client, key=key_1)
    add_setting(client, key=key_2)
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    assert len(rv.json) == 2
    rv = client.get(f'/setting_get_all?key={key_1}')
    assert rv.json[0]['key'] == key_1
    assert len(rv.json) == 1
    rv = client.get(f'/setting_get_all?key={key_2}')
    assert len(rv.json) == 1
    assert rv.json[0]['key'] == key_2


def test_get_all(client):
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    assert len(rv.json) == 0
    key = random_string()
    # test add
    item = add_setting(client, key=key)
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    assert len(rv.json) == 1
    assert rv.json[0]['key'] == key
    with app.app_context():
        # filter by
        rv = Setting.json_filter_by(key=key)
        assert rv.json[0]['key'] == key
    # all as dict list
    result = Setting.query.all()
    dict_list = Setting.dict_list(result)
    assert len(dict_list) == 1
    assert dict_list[0]['key'] == key


def test_get_user(client):
    key = random_string()
    # test add 2 settings
    client.post('/setting_add', data=dict(setting_type='test', key=key, value='test-value'))
    test_user_name = random_string()
    client.post('/setting_add', data=dict(setting_type='test', key=key, value='test-value', user=test_user_name))
    rv = client.get('/setting_user/{}'.format(test_user_name))
    assert len(rv.json) == 1
    item = rv.json[0]
    assert item['user'] == test_user_name
    rv = client.get('/setting_id_user/{}/{}'.format(item['id'], test_user_name))
    assert rv.json['user'] == test_user_name
    # should not be found
    rv = client.get('/setting_id_user/{}/{}'.format(item['id'], random_string()))
    assert rv.status_code == 404
    # get all by user
    rv = client.get('/setting_user/{}'.format('no-one'))
    assert len(rv.json) == 0
    user_404 = None
    try:
        user_404 = Setting.get_by_user_or_404(item_id=item['id'], user='not-here')
    except Exception as e:
        assert '404 Not Found:' in str(e)
    assert user_404 is None


def test_get_property(client):
    key = random_string()
    test_value = random_string()
    flong_value = random_string()
    # test add a thing
    add_setting(client, key=key, value=test_value, number=2)
    setting = Setting.query.filter_by(key=key).first()
    setting.add_sub(flong_value)
    rv = client.get('/setting_get_key/{}'.format(key))
    assert rv.status_code == 200
    assert rv.json['prop_test'] == 'prop:' + test_value
    assert rv.json['prop_datetime'] == str(setting.created).split('.')[0]
    assert rv.json['prop_test_dict'] == {'prop': test_value}
    assert rv.json['prop_complex'] == str(complex(setting.number, setting.number * 2))
    assert set(rv.json['prop_set']) == set([3, 1, 2, 'four'])
    assert rv.json['last_sub_setting']['flong'] == flong_value


def test_relationships(client):
    key = random_string()
    # test add
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value='test-value'))
    # add relation
    setting = Setting.query.filter_by(setting_type='test').first()
    client.post(f'/sub_setting_add/{setting.id}', data=dict(flong='blong', setting=setting))
    # see if returned
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    assert len(rv.json) == 1
    assert len(rv.json[0]['sub_settings']) == 1
    assert rv.json[0]['sub_settings'][0]['flong'] == 'blong'


def test_prop_filters(client):
    # test add
    filter_key = random_string()
    for _ in range(10):
        client.post('/setting_add', data=dict(setting_type='test', key=random_string(), value=random_string()))

    client.post('/setting_add', data=dict(setting_type='test', key=filter_key, value=random_string()))
    rv = client.get(f'/setting_get_all?key={filter_key}')
    assert 200 == rv.status_code
    assert 1 == len(rv.json)
    assert filter_key == rv.json[0]['key']


def test_private_field(client):
    # create
    excluded_key = 'private'
    value = '123789'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=excluded_key, value=value))

    # value is excluded from dict list
    query = Setting.query.filter_by(setting_type='test')
    items = Setting.dict_list(query)
    assert 'key' not in items[0]


def test_can_access_update(client):
    # create
    excluded_key = random_string()
    value = '123456789'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=excluded_key, value=value))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=excluded_key).first()
    excluded_id = item.id
    key = random_string()
    value = random_string()
    client.post('/setting_add', data=dict(setting_type='test', key=key, value=value, user='Robert'))
    key = random_string()
    value = random_string()
    client.post('/setting_add', data=dict(setting_type='test', key=key, value=value))

    # one value is excluded
    with app.app_context():
        r = Setting.json_first(key=excluded_key)
        assert r.status_code == 200
        assert r.response == [b'{}\n']
    # value is excluded from dict list
    query = Setting.query.filter_by(setting_type='test')
    items = Setting.dict_list(query)
    assert len(items) == 2
    with app.app_context():
        r = Setting.json_list(query)
        l = json.loads(r.response[0])
        assert len(l) == 2
    # try to update
    rv = client.put('/setting_update/{}'.format(excluded_id), json=dict(active=False))
    assert rv.status_code != 200
    assert b'Error updating item: Update not allowed.  Magic value!' == rv.data
    result_list = Setting.query_by_access(setting_type='test')
    assert len(result_list) == 2
    result_list = Setting.query_by_access(user='Andrew', setting_type='test')
    assert len(result_list) == 1  # no robert


def test_can_delete(client):
    # create
    key = random_string()
    value = '1234'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == value
    rv = client.delete('/setting_delete/{}'.format(item.id))
    assert rv.status_code == 400
    assert rv.data == b'Deletion not allowed.  Magic value!'


def test_column_conversion(client):
    # create
    key = random_string()
    value = '12.34'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, lob=value))
    # get
    item = Setting.query.filter_by(key=key).first()
    assert item.as_dict['lob'] == '12.34'


def test_update_create_type_conversion(client):
    # create
    key = random_string()
    value = '1234'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    # default bool type conversion
    # json
    rv = client.put('/setting_update/{}'.format(item.id), json=dict(active=False))
    assert rv.status_code == 200, rv.data
    assert b'Updated' == rv.data, rv.data
    item = Setting.query.filter_by(key=key).first()
    assert 'n' == item.active, item
    # query parameters as strings so bool converter does not work
    rv = client.put('/setting_update/{}'.format(item.id), query_string=dict(active='y'))
    assert rv.status_code == 200
    assert rv.data == b'Updated'

    item = Setting.query.filter_by(key=key).first()
    assert 'y' == item.active
    flong = random_string()
    rv = client.post('/sub_setting_add/{}'.format(item.id), data=dict(flong=flong))
    item = SubSetting.query.filter_by(flong=flong).first()
    ss_id = item.id
    ss = client.get('/sub_setting_get/{}'.format(ss_id))
    assert ss.json['id'] == ss_id
    # bool type conversion
    for new_boolean in [True, False]:
        rv = client.put('/sub_setting_put/{}'.format(ss_id), json=dict(boolean=new_boolean))
        assert 200 == rv.status_code, rv
        ss = SubSetting.query.get_or_404(ss_id)
        assert new_boolean == ss.boolean


def test_convert_type(client):
    # add conversion type
    key = random_string()
    value = random_string()
    client.post('/setting_add', data=dict(setting_type='test', key=key, value=value, floaty=123.456))
    item = Setting.query.filter_by(key=key).first()
    float_value_to_convert = 23.4
    int_value_to_convert = 45
    convert_multiple = 2
    # old_convert_type = Setting.convert_types
    # Setting.convert_types = [{'type': int, 'method': lambda n: int(n) * 2},
    #                          {'type': float, 'method': lambda n: float(n) * 2}]
    rv = client.put('/setting_update/{}'.format(item.id),
                    json=dict(number=int_value_to_convert, floaty=float_value_to_convert))
    # Setting.convert_types = old_convert_type
    assert rv.status_code == 200
    assert rv.data == b'Updated'
    item = Setting.query.filter_by(key=key).first()
    # explicit double conversion type
    assert int_value_to_convert * convert_multiple == item.number


def test_sqlite_datetime_convert_type(client):
    # test sqlite conversion types
    date_value = '2004-05-23'
    rv = client.post('/datetest', data=dict(a_date=date_value))
    assert rv.status_code == 200, rv.data
    item = DateTest.query.first()
    assert item.a_date == datetime.strptime(date_value, '%Y-%m-%d')
    date_value = '2010-05-23'
    rv = client.put('/datetest/{}'.format(item.id),
                    json=dict(a_date=date_value))
    assert rv.status_code == 200, rv.data
    item = DateTest.query.first()
    # explicit double conversion type
    assert item.a_date == datetime.strptime(date_value, '%Y-%m-%d')


def test_excluded(client):
    # create
    key = random_string()
    value = '1234'
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    # test non-serialize fields excluded
    rv = client.get('/setting_get/{}'.format(item.id))
    assert 'created' not in rv.json
    assert 'updated' not in rv.json
    assert rv.status_code == 200
    assert 'created' not in item.as_dict
    assert 'updated' in item.as_dict


def test_before_update(client):
    key = random_string()
    test_value = random_string()
    # create using post
    rv = client.post('/setting_post', data=dict(setting_type='test', key=key, value=test_value, number=10))
    assert rv.status_code == 200
    item = Setting.query.get_or_404(rv.json['id'])
    assert item
    assert rv.json['value'] == test_value
    assert item.value == test_value
    assert item.number == 20
    # update using post - missing active should become 'n' via before_update hook
    new_value = random_string()
    rv = client.post('/setting_post/{}'.format(item.id),
                     data=dict(setting_type='test', key=key, value=new_value, number=10))
    assert rv.status_code == 200
    assert rv.json['message'] == 'Updated'
    item = Setting.query.get_or_404(item.id)
    assert 'n' == item.active


def test_get_delete_put_post(client):
    key = random_string()
    # create using post
    rv = client.post('/setting_post', data=dict(setting_type='test', key=key, value='test-value', number=10))
    assert rv.status_code == 200
    item = Setting.query.get_or_404(rv.json['id'])
    assert item
    assert rv.json['value'] == 'test-value'
    assert item.value == 'test-value'
    assert item.number == 20
    # update using post
    new_value = random_string()
    rv = client.post('/setting_post/{}'.format(item.id),
                     data=dict(setting_type='test', key=key, value=new_value, number=10))
    assert rv.status_code == 200
    assert rv.json['message'] == 'Updated'
    # test update_properties are returned
    assert rv.json['properties']['prop_test'] == 'prop:' + new_value
    assert rv.json['item']['key'] == key
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == new_value
    assert 20 == item.number
    # post item not found
    rv = client.post('/setting_post/{}'.format(random.randint(100, 999)),
                     data=dict(setting_type='test', key=key, value='new-value', number=10))
    assert 404 == rv.status_code
    # put not valid in something meant for post
    rv = client.put('/setting_post')
    assert 400 == rv.status_code
    # post fail validation
    rv = client.post('/setting_post/{}'.format(item.id), data=dict(key=''))
    assert 400 == rv.status_code
    assert rv.data == b'Missing key'
    # update using put
    new_value = random_string()
    new_number = random.randint(0, 999)
    rv = client.put('/setting_put/{}'.format(item.id),
                    json=dict(setting_type='test', key=key, value=new_value, number=new_number))
    assert rv.status_code == 200
    assert rv.json['message'] == 'Updated'
    assert rv.json['properties']['prop_test'] == 'prop:' + new_value
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == new_value
    assert item.number == new_number * 2
    # put fail validation
    rv = client.put('/setting_put/{}'.format(item.id), data=dict(key=''))
    print(rv)
    assert rv.status_code == 400
    assert rv.data == b'Missing key'
    # create post fails
    rv = client.post('/setting_post', data=dict(flong='fling'))
    assert rv.status_code == 400
    assert rv.data == b'Missing key'


def test_create_update_json(client):
    # create
    key = random_string()
    value = random_string()
    rv = client.post('/setting_post', json=dict(setting_type='test', key=key, value=value, number=10))
    assert rv.status_code == 200
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == value
    value = random_string()
    dt_now = datetime.utcnow()
    rv = client.post(f'/setting_post/{item.id}',
                     json=dict(setting_type='test', key=key, value=value, number=10,
                               scheduled=dt_now.strftime(Setting.scheduled_date_format)))
    assert rv.status_code == 200, rv.data
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == value
    assert item.scheduled.strftime(Setting.scheduled_date_format) == dt_now.strftime(Setting.scheduled_date_format)


def test_create_update_delete(client):
    # create
    key = random_string()
    value = random_string()
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value, number=10))
    assert rv.status_code == 302, rv.data
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == value
    # test that number is not set on creation as it is not included in create_fields
    assert item.number == 20
    old_updated = item.updated

    new_value = random_string()
    item.update_from_dict(dict(value=new_value))
    assert item.previous_field_value['value'] == value
    assert item.value == new_value
    # set to new value
    new_value = random_string()
    rv = client.post('/setting_edit/{}'.format(item.id),
                     data=dict(value=new_value, number=100))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == new_value
    assert item.number == 200
    # check updated is changing
    assert old_updated != item.updated
    # set to ''
    rv = client.post('/setting_edit/{}'.format(item.id),
                     data=dict(value=''))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert not item.value
    # fail validation
    rv = client.post('/setting_edit/{}'.format(item.id),
                     data=dict(key=''))
    assert rv.status_code == 500
    assert b'Error updating item: Missing key' in rv.data
    # delete
    rv = client.delete('/setting_delete/{}'.format(item.id))
    assert rv.status_code == 200
    assert rv.json['item']['id'] == item.id
    item = Setting.query.filter_by(key=key).first()
    assert not item


def test_form_page(client):
    # create
    key = random_string()
    rv = client.post('/setting_form_add', data=dict(setting_type='test', key=key, value='test-value', number=10))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == 'test-value'
    # test that number is not set on creation as it is not included in create_fields
    assert item.number == 20
    # set to new value
    new_value = random_string()
    rv = client.post('/setting_form_edit/{}'.format(item.id),
                     data=dict(id=item.id, setting_type='test', key=key, value=new_value, number=100))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first()
    assert item
    assert item.value == new_value
    assert item.number == 200
    # set to ''
    rv = client.post('/setting_form_edit/{}'.format(item.id),
                     data=dict(id=item.id, setting_type='test', key=key, value=''))
    assert rv.status_code == 302
    item = Setting.query.filter_by(key=key).first_or_404()
    assert not item.value
    # fail validation
    rv = client.post('/setting_form_edit/{}'.format(item.id),
                     data=dict(id=item.id, setting_type='test', key=''))
    assert rv.status_code == 200
    assert b'Missing key' in rv.data


def test_default_create_fields(client):
    key = random_string()
    value = random_string()
    prop_test = random_string()
    # add
    old_create_fields = Setting.create_fields
    # remove fields
    Setting.create_fields = []
    rv = client.post('/setting_add', data=dict(setting_type='test', key=key, value=value, prop_test=prop_test))
    assert 302 == rv.status_code, rv.data
    # assert it was created
    item = Setting.query.filter_by(key=key).first()
    assert item.key == key, item
    assert item.value == value, item
    assert item.prop_test == 'prop:' + value, item
    Setting.create_fields = old_create_fields


def test_simple_model_update_fields(client):
    value = random_string()
    # add
    item = SimpleModel(value=value)
    db.session.add(item)
    db.session.commit()
    item = SimpleModel.query.filter_by(value=value).first_or_404()
    # form
    value = random_string()
    prop = random_string()
    rv = client.post('/simple_edit/{}'.format(item.id), data=dict(value=value, prop=prop))
    assert 200 == rv.status_code, rv.data
    assert rv.json['item']['value'] == value, rv.data
    assert b'Updated' in rv.data, rv.data
    # json
    value = random_string()
    rv = client.put('/simple_edit/{}'.format(item.id), json=dict(value=value, id='dskdsf'))
    assert 200 == rv.status_code, rv.data
    assert rv.json['item']['value'] == value, rv.data
    assert rv.json['item']['prop'] == 'prop:' + value, rv.data
    assert b'Updated' in rv.data, rv.data


def test_override_datetime_conversion(client):
    key = random_string()
    test_value = random_string()
    # test add a thing
    item = add_setting(client, key=key, value=test_value)
    sub_setting = SubSetting(setting=item, flong=random_string(5))
    db.session.add(sub_setting)
    db.session.commit()
    item = Setting.query.filter_by(key=key).first()
    sub = item.sub_settings[0]
    unix_time = int(time.mktime(sub.created.timetuple())) * 1000
    assert type(sub.as_dict['created']) == int
    assert sub.as_dict['created'] == unix_time


def test_json_get(client):
    key = random_string()
    test_value = random_string()
    setting_type = random_string()
    # test add setting
    Setting.__model_props = {}
    item = Setting(setting_type=setting_type, key=key, value=test_value)
    db.session.add(item)
    db.session.commit()
    item = Setting.query.get_or_404(item.id)
    # get by id
    rv = client.get('/setting_get_json/{}'.format(item.id))
    assert rv.json['value'] == test_value
    assert rv.json['key'] == key
    assert rv.json['setting_type'] == setting_type
    rv = client.get('/setting_get_json/{}'.format(item.id + 100))
    assert rv.status_code == 200
    assert rv.json == {}
    # get first
    rv = client.get('/setting_json_first/{}'.format(item.key))
    assert rv.json['value'] == test_value
    assert rv.json['key'] == key
    assert rv.json['setting_type'] == setting_type
    assert client.get('/setting_json_first/{}'.format(random_string())).json == {}


def test_get_0_is_not_null(client):
    key = random_string()
    item = Setting(id=0, setting_type='hello', value=random_string(), key=key)
    db.session.add(item)
    db.session.commit()
    # should get a list
    rv = client.get('/setting_get_all')
    assert rv.status_code == 200
    assert len(rv.json) == 1
    assert rv.json[0]['key'] == key
    # should get one item not a list
    rv = client.get('/setting_get/0')
    assert rv.status_code == 200
    assert rv.json['key'] == key


def test_timestamp_is_updated_and_can_be_overridden(client):
    key = random_string()
    item = Setting(setting_type='hello', value=random_string(), key=key)
    db.session.add(item)
    db.session.commit()
    new_value = random_string()
    item = Setting.query.get_or_404(item.id)
    sub_item = item.add_sub(new_value)
    sub_item_id = sub_item.id
    updated_when_created = sub_item.sub_updated
    # update using put
    new_value = random_string()
    assert 200 == client.put('/sub_setting_put/{}'.format(sub_item_id), json=dict(flong=new_value)).status_code
    updated_item = SubSetting.query.get_or_404(sub_item_id)
    assert updated_item.flong == new_value
    # test custom update works and that timestamp_fields works
    assert updated_when_created > updated_item.sub_updated
